#include "matrix.h"

/*
 * Retourne une nouvelle matrice
 * w : largeur
 * h : hauteur
 */
matrix * matrix_new(int w, int h)
{
    assert(w > 0);
    assert(h > 0);

    int i;
    matrix * m = malloc(sizeof(matrix));
    
    if ( ! m )
	return NULL;
    
    m->w = w;
    m->h = h;
    
    m->matrix = calloc(h, sizeof(int *));
    if ( ! m->matrix ) {
	free(m);
	return NULL;
    }

    for ( i=0 ; i < m->h ; i++ ) {
	m->matrix[i] = calloc(m->w, sizeof(int));
	if ( ! m->matrix[i] ) {
	    while ( i > 0 ) {
		i--;
		free(m->matrix[i]);
	    }
	    free(m->matrix);
	    free(m);
	    return NULL;
	}
    }

    return m;
}

/*
 * Retourne une matrice de controle de Hamming
 */
matrix * matrix_h(int n)
{
    int i,j;
    matrix * h = matrix_new((1<<n)-1, n);
    
    if ( ! h )
	return NULL;

    for ( i=0 ; i < h->w ; i++ )
	for ( j=0 ; j < n ; j++ )
	    h->matrix[j][i] = ((i+1) & (1<<(n-j-1)))?1:0;

    return h;
}

/*
 * Retourne une matrice de codage de Hamming
 */
matrix * matrix_g(int n)
{
    if ( n < 2 ) {
	fprintf(stderr,"Impossible de calculer une matrice G pour n < 2\n");
	return NULL;
    }

    int i,j,
	c, // En cours de dcomposition
	d = 3, // Nombre  dcomposer
	ppdd = 4; // Prochaine Puissance De Deux
    matrix * g = matrix_new((1<<n)-1, (1<<n)-n-1);

    if ( ! g )
	return NULL;

    for ( i=0 ; i < g->h ; i++ ) {
	if ( d == ppdd ) {
	    d++;
	    ppdd <<= 1;
	}
	c = d;
	g->matrix[i][d-1] = 1;
	for ( j=d-1 ; j >= 0 ; j-- ) {
	    if ( c-(1<<j) >= 0 && c > 0 ) {
		g->matrix[i][(1<<j)-1] = 1;
		c -= 1<<j;
	    }
	}
	d++;
    }

    return g;
}

/*
 * Fait le produit binaire d'un vecteur ligne et d'une matrice
 */
matrix * matrix_prodLM(matrix * l, matrix * m)
{
    assert(l->w == m->h);
    assert(l->h == 1);

    int i,j;
    matrix * prod = matrix_new(m->w, 1);

    if ( ! prod )
	return NULL;

    for ( i=0 ; i < l->w ; i++ )
	if ( l->matrix[0][i] )
	    for ( j=0 ; j < m->w ; j++ )
		prod->matrix[0][j] ^= m->matrix[i][j];

    return prod;
}

/*
 * Fait le produit binaire d'une matrice et d'un vecteur colonne
 */
matrix * matrix_prodMC(matrix * m, matrix * c)
{
    assert(m->w == c->h);
    assert(c->w == 1);

    matrix * prod = matrix_new(m->h, 1);

    if ( ! prod )
	return NULL;

    int i,j;
    for ( i=0 ; i < m->h ; i++ )
	for ( j=0 ; j < m->w ; j++ )
	    if ( m->matrix[i][j] )
		prod->matrix[0][i] ^= c->matrix[j][0];

    return prod;
}

/*
 * Teste si une matrice est remplie de 0
 */
int matrix_null(matrix * m)
{
    assert(m != NULL);

    int i,j;

    for ( i=0 ; i < m->h ; i++ )
	for ( j=0 ; j < m->w ; j++ )
	    if ( m->matrix[i][j] )
		return 0;

    return 1;
}

/*
 * Libre la matrice donne
 */
void matrix_free(matrix * m)
{
    while ( m->h > 0 ) {
	m->h--;
	free(m->matrix[m->h]);
    }
    free(m->matrix);
    free(m);
}

/*
 * Retourne l'exposant donnant de la puissance de 2 qui donne n
 */
static int power(int n)
{
    int count=-1;
    while ( n != 0 ) {
	count++;
	n >>= 1;
    }
    return count;
}

static int isPowerOfTwo(int n)
{
    int i, count;
    for ( i=0, count=0 ; i != 32 ; i++ ) {
	if ( (1<<i) & n )
	    count++;
	if ( count > 1 )
	    return 0;
    }
    return 1;
}

/*
 * Retourne un vecteur ligne recopis sur le vecteur colonne m priv de ces bits
 * dont la position est une puissance de 2
 */
matrix * matrix_databits(matrix * m)
{
    assert(m->w == 1);

    matrix * data = matrix_new(m->h - power(m->h+1), 1);
    int i,j=0;

    for ( i=0 ; i < m->h ; i++ )
	if ( ! isPowerOfTwo(i+1) ) {
	    data->matrix[0][j] = m->matrix[i][0];
	    j++;
	}
    
    return data;
}

/*
 * Affiche la matrice m
 */
void matrix_print(matrix * m)
{
    int i,j;
    for ( i=0 ; i < m->h ; i++ ) {
	for ( j=0 ; j < m->w ; j++ )
	    printf("%d",m->matrix[i][j]);
	putchar('\n');
    }
}
